﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using STFXPdfDotNET;
using System.IO;
using System.Windows.Forms;
using Ionic.Zip;
using STFXmlParser;
using System.Drawing;

namespace CalcHelper
{
    public class ProcessPdf
    {
        private string _extPath = null;

        private int _vertScrollW = System.Windows.Forms.SystemInformation.VerticalScrollBarWidth;
        private int _horScrollH = System.Windows.Forms.SystemInformation.HorizontalScrollBarHeight;

        public double zoomOutVal = 0.8;
        public double zoomInVal = 1.25;
        public int zoomCntMax = 5;
        public bool initZoomReset = false;
        public bool curPageZoomDelta = false;


        public int zoomCnt = 0;

        public int vScrollPos = 0;
        public int hScrollPos = 0;
        public Boolean centerPic = true;
        public Boolean resetScrollBars = true;

        public Point mDown;
        public Boolean mLassoDown = false;
        public Boolean mLassoMove = false;
        public Rectangle lassoRect;

        public int boxedPageIndex;
        public uint numPages;
        public uint curPage;
        public List<uint> boxedPageList;
        public Boolean pageNav = true;

        public string pdfFile = null;
        public string xmlFile = null;
        public string id = "";

        private Bitmap _bMapBase = null;
        public Bitmap bMap = null;

        public Panel panel;
        public PictureBox pictureBox;

        public ProcessPdf(string pdfId, ref Panel pnl, ref PictureBox picBox)
        {
            id = pdfId;
            panel = pnl;
            pictureBox = picBox;
        }

        public string ExtPath
        {
            get { return _extPath; }
            set { _extPath = value; }
        }

        public void ShowPage(int page, Boolean newBase, ref PageArray pa)
        {
            CSTFPdf STFPdf = new CSTFPdf();
            PdfHandle hPdf = null;
            RectangleF[] rect = new RectangleF[1];
            Pen pColor = new Pen(Color.Blue, 1);
            SolidBrush tBrush = new SolidBrush(Color.Blue);
            List<Field> fa = null;
            FontStyle fontStyle;
            StringFormat drawFormat = new StringFormat(StringFormat.GenericTypographic);
            PageScale ps;
            Font font;
            Boolean isChk = false;
            Boolean isMulti = false;
            Boolean rotateCid = false;
            double dpi = 0;
            string fontName = "";
            string refName = "";
            char[] check = { (char)0xfc };
            char[] multi = { (char)0x32 };
            float fontSize = 0.0f;

            if (!string.IsNullOrEmpty(pdfFile))
            {
                try
                {
                    // Get the PageScale info for the current page:
                    ps = pa.GetPageScale(page - 1);

                    if (ps == null)
                    {
                        MessageBox.Show(string.Format("pa.GetPageScale({0}) has returned a null ScaleField.", page - 1), "ShowPage():", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        return;
                    }

                    if (newBase)
                    {
                        // Set the bitmap base.
                        hPdf = STFPdf.OpenPdf(pdfFile, "", "", ShareType.ShareTypeNone);
                        dpi = ps.dpi;

                        if (_bMapBase != null)
                        {
                            _bMapBase.Dispose();
                        }

                        _bMapBase = STFPdf.ConvertPageToBitmap(hPdf, (uint)page, dpi, false, false, true);
                    }

                    if (bMap != null)
                    {
                        bMap.Dispose();
                    }

                    bMap = new Bitmap(_bMapBase);

                    pa.GetFieldArrayRef(page - 1, ref fa);

                    if (fa != null)
                    {
                        if (ps.yScale == 0.0f)
                        {
                            MessageBox.Show(string.Format("pa.GetPageScale({0}) has returned a yScale of 0.0.", page - 1), "ShowPage():", MessageBoxButtons.OK, MessageBoxIcon.Error);
                            return;
                        }

                        drawFormat.FormatFlags = StringFormatFlags.NoClip | StringFormatFlags.NoWrap;
                        drawFormat.LineAlignment = StringAlignment.Far;
                        drawFormat.Trimming = StringTrimming.None;

                        using (Graphics gr = Graphics.FromImage(bMap))
                        {
                            gr.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAlias;

                            for (int i = 0; i < fa.Count; i++)
                            {
                                Field field = fa[i];

                                // box1_65535 is a special CID related box which we will not bother with.
                                // We are currently drawing the associated rectangle whose coords are all 0.0.
                                if (field.ref_name == "box1_65535")
                                {
                                    continue;
                                }

                                fontStyle = FontStyle.Regular;

                                if (field is BarcodeField)
                                {
                                    pColor.Color = Color.FromArgb(255, 128, 0);
                                    tBrush.Color = Color.FromArgb(255, 128, 0);
                                    drawFormat.Alignment = StringAlignment.Near;
                                    fontName = "Courier New";
                                    fontSize = 12.0f * (float)ps.yScale;
                                    refName = field.ref_name.Substring(3);
                                }
                                else if (field is TextField)
                                {
                                    isChk = ((TextField)field).checkbox;
                                    isMulti = ((TextField)field).multiline;
                                    fontName = ((TextField)field).font_name;
                                    fontSize = (float)(((TextField)field).font_size) * (float)ps.yScale;
                                    refName = field.ref_name.Substring(3);

                                    if (((TextField)field).font_bold)
                                    {
                                        fontStyle = FontStyle.Bold;
                                    }

                                    if (((TextField)field).font_italic)
                                    {
                                        fontStyle |= FontStyle.Italic;
                                    }

                                    switch (((TextField)field).just)
                                    {
                                        case (TextField.justification.left):

                                            pColor.Color = Color.Blue;
                                            tBrush.Color = Color.Blue;
                                            drawFormat.Alignment = StringAlignment.Near;
                                            break;

                                        case (TextField.justification.right):

                                            pColor.Color = Color.Red;
                                            tBrush.Color = Color.Red;
                                            drawFormat.Alignment = StringAlignment.Far;
                                            break;

                                        case (TextField.justification.center):

                                            pColor.Color = Color.Magenta;
                                            tBrush.Color = Color.Magenta;
                                            drawFormat.Alignment = StringAlignment.Center;
                                            break;

                                        case (TextField.justification.dec):

                                            pColor.Color = Color.FromArgb(0, 192, 0);
                                            tBrush.Color = Color.FromArgb(0, 192, 0);
                                            drawFormat.Alignment = StringAlignment.Far;
                                            break;

                                        case (TextField.justification.dec_np):

                                            pColor.Color = Color.FromArgb(0, 100, 0);
                                            tBrush.Color = Color.FromArgb(0, 100, 0);
                                            drawFormat.Alignment = StringAlignment.Far;
                                            break;
                                    }
                                }
                                else if (field is CidField)
                                {
                                    fontName = ((CidField)field).font_name;
                                    fontSize = (float)(((CidField)field).font_size) * (float)ps.yScale;
                                    refName = field.ref_name;

                                    if (((CidField)field).font_bold)
                                    {
                                        fontStyle = FontStyle.Bold;
                                    }

                                    if (((CidField)field).font_italic)
                                    {
                                        fontStyle |= FontStyle.Italic;
                                    }

                                    switch (((CidField)field).just)
                                    {
                                        case (CidField.justification.left):

                                            pColor.Color = Color.Blue;
                                            tBrush.Color = Color.Blue;
                                            drawFormat.Alignment = StringAlignment.Near;
                                            break;

                                        case (CidField.justification.right):

                                            pColor.Color = Color.Red;
                                            tBrush.Color = Color.Red;
                                            drawFormat.Alignment = StringAlignment.Far;
                                            break;

                                        case (CidField.justification.center):

                                            pColor.Color = Color.Magenta;
                                            tBrush.Color = Color.Magenta;
                                            drawFormat.Alignment = StringAlignment.Center;
                                            break;
                                    }

                                    if (((CidField)field).rotation != "0")
                                    {
                                        rotateCid = true;
                                    }
                                }

                                font = new Font(fontName, fontSize, fontStyle, GraphicsUnit.Pixel);
                                rect[0] = field.sbox;
                                gr.DrawRectangles(pColor, rect);

                                // Need to check for CID rotation before we draw the text.
                                if (rotateCid)
                                {
                                    rotateCid = false;

                                    PointF pt = new Point();

                                    // Rotation works clockwise.
                                    switch (((CidField)field).rotation)
                                    {
                                        case ("90"):

                                            pt.X = field.sbox.X + field.sbox.Width;

                                            switch (((CidField)field).just)
                                            {
                                                case (CidField.justification.left):

                                                    pt.Y = field.sbox.Y + field.sbox.Height;
                                                    break;

                                                case (CidField.justification.center):

                                                    pt.Y = field.sbox.Y + field.sbox.Height / 2.0f;
                                                    break;

                                                case (CidField.justification.right):

                                                    pt.Y = field.sbox.Y;
                                                    break;
                                            }

                                            gr.TranslateTransform(pt.X, pt.Y);
                                            gr.RotateTransform(270);

                                            break;

                                        case ("180"):

                                            pt.Y = field.sbox.Y;

                                            switch (((CidField)field).just)
                                            {
                                                case (CidField.justification.left):

                                                    pt.X = field.sbox.X + field.sbox.Width;
                                                    break;

                                                case (CidField.justification.center):

                                                    pt.X = field.sbox.X + field.sbox.Width / 2.0f;
                                                    break;

                                                case (CidField.justification.right):

                                                    pt.X = field.sbox.X;
                                                    break;
                                            }

                                            gr.TranslateTransform(pt.X, pt.Y);
                                            gr.RotateTransform(180);

                                            break;

                                        case ("270"):

                                            pt.X = field.sbox.X;

                                            switch (((CidField)field).just)
                                            {
                                                case (CidField.justification.left):

                                                    pt.Y = field.sbox.Y;
                                                    break;

                                                case (CidField.justification.center):

                                                    pt.Y = field.sbox.Y + field.sbox.Height / 2.0f;
                                                    break;

                                                case (CidField.justification.right):

                                                    pt.Y = field.sbox.Y + field.sbox.Height;
                                                    break;
                                            }

                                            gr.TranslateTransform(pt.X, pt.Y);
                                            gr.RotateTransform(90);

                                            break;
                                    }

                                    gr.DrawString(refName, font, tBrush, 0, 0, drawFormat);
                                    gr.ResetTransform();

                                }
                                else
                                {
                                    gr.DrawString(refName, font, tBrush, field.sbox, drawFormat);
                                }

                                // Impossible to make a box both a checkbox AND a multiline box.
                                if (isChk || isMulti)
                                {
                                    fontStyle = FontStyle.Regular;
                                    font = new Font("wingdings", fontSize, fontStyle, GraphicsUnit.Pixel);

                                    if (isChk)
                                    {
                                        refName = new string(check);
                                        isChk = false;
                                    }
                                    else
                                    {
                                        refName = new string(multi);
                                        isMulti = false;
                                    }

                                    gr.DrawString(refName, font, tBrush, field.sbox, drawFormat);
                                }

                                // Is box selected?
                                if (field.isSelected)
                                {
                                    using (Pen penSel = new Pen(Color.Black, 3f))
                                    {
                                        //penSel.DashStyle = System.Drawing.Drawing2D.DashStyle.Custom;
                                        //penSel.DashPattern = new float[] { 0.5F, 1.0F };
                                        penSel.Alignment = System.Drawing.Drawing2D.PenAlignment.Outset;
                                        penSel.DashStyle = System.Drawing.Drawing2D.DashStyle.Dot;
                                        gr.DrawRectangles(penSel, rect);
                                    }
                                }
                            }
                        }
                    }

                    if (pictureBox.Image != null)
                    {
                        pictureBox.Image.Dispose();
                    }

                    pictureBox.Image = bMap;

                    if (centerPic)
                    {
                        CenterPicPanel();
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.ToString(), "ShowPage():", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                finally
                {
                    if (hPdf != null)
                    {
                        STFPdf.ClosePdf(hPdf);
                    }
                }
            }
        }

        public Boolean ScaleRects(string mode, ref PageArray pa)
        {
            Boolean success = true;
            CSTFPdf STFPdf = new CSTFPdf();
            PdfHandle hPdf = null;
            List<Field> fa = null;
            PageData pd;
            PageScale ps = null;
            PgBMapData bm = null;
            PgZoomData zd = null;
            double pgBmapH = 0, pgBmapW = 0, dpi = 0;
            double xScale, yScale;
            double stfPageH = 0, stfPageW = 0;
            double wDiff, hDiff;
            int stfPageRot;
            float x, y, width, height;

            if (pa == null)
            {
                MessageBox.Show("ScaleRects() received a null pa argument.", "null PageArray argument:", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            if (!string.IsNullOrEmpty(pdfFile))
            {
                try
                {
                    if (mode == "OpenFile")
                    {
                        hPdf = STFPdf.OpenPdf(pdfFile, "", "", ShareType.ShareTypeNone);
                        numPages = STFPdf.GetPageCount(hPdf);
                    }

                    for (uint page = 1; page <= numPages; page++)
                    {
                        if (mode != "OpenFile")
                        {
                            // Get the PageData info for the current page:
                            pd = pa.GetPageData((int)(page - 1));

                            if (pd == null)
                            {
                                MessageBox.Show(string.Format("pa.GetPageData({0}) has returned a null PageData.", page - 1), string.Format("ScaleRects({0}):", mode), MessageBoxButtons.OK, MessageBoxIcon.Error);
                                return false;
                            }
                            else
                            {
                                stfPageH = pd.pageH;
                                stfPageW = pd.pageW;
                                stfPageRot = pd.pageRot;
                            }

                            if ((mode == "ZoomIn") || (mode == "ZoomOut") || (mode == "ZoomReset") || (mode == "Resize"))
                            {
                                // Consider continuing on to next page instead of returning 'false' if
                                // trouble retrieving ps and bm data.

                                if (mode != "Resize")
                                {
                                    // Get the PageScale info for the current page:
                                    ps = pa.GetPageScale((int)(page - 1));

                                    if (ps == null)
                                    {
                                        MessageBox.Show(string.Format("pa.GetPageScale({0}) has returned a null ScaleField.", page - 1), string.Format("ScaleRects({0}):", mode), MessageBoxButtons.OK, MessageBoxIcon.Error);
                                        return false;
                                    }
                                }

                                // Get the Page Bitmap Data for the current page:
                                bm = pa.GetPgBMapData((int)(page - 1));

                                if (bm == null)
                                {
                                    MessageBox.Show(string.Format("pa.GetPgBMapData({0}) has returned a null PgBMapData.", page - 1), string.Format("ScaleRects({0}):", mode), MessageBoxButtons.OK, MessageBoxIcon.Error);
                                    return false;
                                }
                            }
                        }

                        switch (mode)
                        {
                            // The 2 down below is the panel border width:
                            case ("OpenFile"):

                                stfPageH = (double)STFPdf.GetPageHeight(hPdf, page);
                                stfPageW = (double)STFPdf.GetPageWidth(hPdf, page);
                                stfPageRot = STFPdf.GetPageRotation(hPdf, page);

                                pa.SetPageData((int)(page - 1), stfPageH, stfPageW, stfPageRot);

                                if ((stfPageRot == 90) || (stfPageRot == 270) || (stfPageW > stfPageH))
                                {
                                    // Landscape:
                                    pgBmapH = (double)(panel.Width - 2);
                                    pgBmapW = (pgBmapH / stfPageH) * stfPageW;

                                    dpi = 72.0 * (pgBmapH / stfPageH);
                                }
                                else
                                {
                                    pgBmapW = (double)(panel.Width - 2);
                                    pgBmapH = (pgBmapW / stfPageW) * stfPageH;

                                    if (pgBmapH > (double)(panel.Height - 2))
                                    {
                                        // A vertical scrollbar will appear when this page is viewed, so adjust dpi for the width.
                                        pgBmapW = (double)(panel.Width - (2 + _vertScrollW));
                                        pgBmapH = (pgBmapW / stfPageW) * stfPageH;
                                    }

                                    dpi = 72.0 * (pgBmapW / stfPageW);
                                }

                                /*
                                 * Added 07/18/17 - Saw a strange situation after adding a split container so I could 
                                 * show two pdf's on the form: At a proper reduced form size, the above logic would end up
                                 * with a pdf page being displayed with visible form area all around the edges. So I added
                                 * this 'resize' code here to take care of that. This way we can scale the pdf and the
                                 * rectangles in one pass per page. The resize is also further down below as a separate
                                 * operation.
                                */
                                CheckFullWidthPage(ref pgBmapW, ref pgBmapH, ref dpi, stfPageW, stfPageH);

                                break;

                            case ("FullWidth"):

                                initZoomReset = true;

                                pgBmapW = (double)(panel.Width - 2);
                                pgBmapH = (pgBmapW / stfPageW) * stfPageH;

                                if (pgBmapH > (double)(panel.Height - 2))
                                {
                                    // A vertical scrollbar will appear when this page is viewed, so adjust dpi for the width.
                                    pgBmapW = (double)(panel.Width - (2 + _vertScrollW));
                                    pgBmapH = (pgBmapW / stfPageW) * stfPageH;
                                }

                                dpi = 72.0 * (pgBmapW / stfPageW);

                                /*
                                 * Same strange situation after adding a split container....
                                */
                                CheckFullWidthPage(ref pgBmapW, ref pgBmapH, ref dpi, stfPageW, stfPageH);

                                break;

                            case ("FullPage"):

                                initZoomReset = true;

                                pgBmapH = (double)(panel.Height - 2);
                                pgBmapW = (pgBmapH / stfPageH) * stfPageW;

                                if (pgBmapW > (double)(panel.Width - 2))
                                {
                                    // A horizontal scrollbar will appear when this page is viewed, so adjust dpi for the height.
                                    pgBmapH = (double)(panel.Height - (2 + _horScrollH));
                                    pgBmapW = (pgBmapH / stfPageH) * stfPageW;
                                }

                                dpi = 72.0 * (pgBmapH / stfPageH);

                                /*
                                 * Added this here too even though I have not seen it, the "FullWidth" is in the same
                                 * family and I saw it there.
                                */
                                CheckFullWidthPage(ref pgBmapW, ref pgBmapH, ref dpi, stfPageW, stfPageH);

                                break;

                            case ("ZoomOut"):

                                /*
                                 * Each page gets zoomed out and has it's own zoom count.
                                 * Doing it this way since we have documents with landscape
                                 * orientation mixed in with regular pages, legal, etc.
                                 * which throws a curve into handling cases where one page
                                 * gets zoomed out at zoomCnt of -2 [no scroll bars] while
                                 * another page with a different orientation is at zoomCnt
                                 * of -4 and still has scrollbars.
                                */
                                int sBarMask = 0;
                                bool dpiAutoDelta = false;

                                pa.GetPgZoomDataRef((int)page - 1, ref zd);

                                if (zd == null)
                                {
                                    MessageBox.Show(string.Format("pa.GetPgZoomDataRef({0}, ref zd) has returned a null PgZoomData\n\nPage will not be Zoomed Out.", page - 1), "ScaleRects(ZoomOut)", MessageBoxButtons.OK, MessageBoxIcon.Error);
                                    continue;
                                }

                                if (zd.curDpiAutoDelta)
                                {
                                    // Already zoomed out as far as we can go, on to the next page.
                                    continue;
                                }
                                else
                                {
                                    /* 
                                     * Note that scrollbars do not appear if bitmap size prevents
                                     * the picture box from expanding 'past' the enclosing panel's edges.
                                     * We are checking the stored bitmap width and height values for this
                                     * page against the panel size, minus the 2 pixel edge widths for
                                     * each dimension. You would have to be viewing the page in order to use
                                     * form.PicPanel.ClientSize values, so we are doing it this way.
                                     * 
                                     * Our picture panel has a 1 pixel boundary the way it is currently set
                                     * in the Designer [so you get 2 for Width and Height].
                                     * 
                                     * Continue on to the next page if there would be no current scrollbars.
                                     */
                                    if (bm.pgBmapH > (double)(panel.Height - 2))
                                    {
                                        // Vertical scrollbar will appear when page is viewed.
                                        sBarMask = 10;
                                    }

                                    if (bm.pgBmapW > (double)(panel.Width - 2))
                                    {
                                        // Horizontal scrollbar will appear when page is viewed.
                                        sBarMask += 1;
                                    }

                                    if (sBarMask > 0)
                                    {
                                        switch (sBarMask)
                                        {
                                            case (1):

                                                // A horizontal scrollbar is/will be present -
                                                // 'is' -> if page is currently being viewed.
                                                pgBmapW = bm.pgBmapW * zoomOutVal;

                                                if (pgBmapW > (double)(panel.Width - 2))
                                                {
                                                    // Horizontal scrollbar will still remain after reduction.
                                                    pgBmapH = bm.pgBmapH * zoomOutVal;

                                                    dpi = ps.dpi * zoomOutVal;
                                                }
                                                else
                                                {
                                                    /*
                                                     * Horizontal scrollbar will disappear after reduction, so
                                                     * perform the "FullWidth" logic [we already know there is no
                                                     * vertical scrollbar]:
                                                    */
                                                    pgBmapW = (double)(panel.Width - 2);
                                                    pgBmapH = (pgBmapW / stfPageW) * stfPageH;

                                                    dpi = 72.0 * (pgBmapW / stfPageW);

                                                    dpiAutoDelta = true;
                                                }

                                                break;

                                            case (10):

                                                // A vertical scrollbar is/will be present -
                                                // 'is' -> if page is currently being viewed.
                                                pgBmapH = bm.pgBmapH * zoomOutVal;

                                                if (pgBmapH > (double)(panel.Height - 2))
                                                {
                                                    // Vertical scrollbar will still remain after reduction.
                                                    pgBmapW = bm.pgBmapW * zoomOutVal;

                                                    dpi = ps.dpi * zoomOutVal;
                                                }
                                                else
                                                {
                                                    /*
                                                     * Vertical scrollbar will disappear after reduction, so
                                                     * perform the "FullPage" logic [we already know there is no
                                                     * horizontal scrollbar]:
                                                    */
                                                    pgBmapH = (double)(panel.Height - 2);
                                                    pgBmapW = (pgBmapH / stfPageH) * stfPageW;

                                                    dpi = 72.0 * (pgBmapH / stfPageH);

                                                    dpiAutoDelta = true;
                                                }

                                                break;

                                            case (11):

                                                /*
                                                 * A vertical AND a horizontal scrollbar is/will be present -
                                                 * 'is' -> if page is currently being viewed.
                                                 * We now consider the scrollbar situation AFTER the reduction
                                                 * math - we will re-use sBarMask.
                                                */
                                                sBarMask = 0;

                                                pgBmapH = bm.pgBmapH * zoomOutVal;
                                                pgBmapW = bm.pgBmapW * zoomOutVal;

                                                if (pgBmapH > (double)(panel.Height - 2))
                                                {
                                                    // Vertical scrollbar will still remain after reduction.
                                                    sBarMask = 10;
                                                }

                                                if (pgBmapW > (double)(panel.Width - 2))
                                                {
                                                    // Horizontal scrollbar will still remain after reduction.
                                                    sBarMask += 1;
                                                }

                                                if (sBarMask > 0)
                                                {
                                                    /*
                                                     * We will either have both or one scrollbars after reduction.
                                                     * We only need to reduce the dpi for this page and move on further
                                                     * below.
                                                    */
                                                    dpi = ps.dpi * zoomOutVal;
                                                }
                                                /* Both scrollbars will be gone after reduction. We need to determine
                                                 * either a "FullWidth" or a "FullPage" reduction.
                                                */
                                                else if (CheckFullWidthPage(ref pgBmapW, ref pgBmapH, ref dpi, stfPageW, stfPageH))
                                                {
                                                    dpiAutoDelta = true;
                                                }
                                                break;
                                        }

                                        /*
                                         * Check if we 'forced' a dpi change:
                                        */
                                        if (dpiAutoDelta)
                                        {
                                            /*
                                             * zd.curDpiAutoDelta is set when we hit the bottom end of
                                             * our zoom out, that is, we can zoom out no more because we
                                             * ran out of scrollbars. A zoom in will reset it.
                                             * 
                                             * zd.mstrDpiAutoDelta is used to indicate that we hit
                                             * the lower end of the zoom out for this page and we will 
                                             * then use the dpi and bitmap values stored in zd when zooming
                                             * in [up] to a zd.zoomCnt of zero. zd.mstrDpiAutoDelta will be
                                             * reset when we set the zd.zoomCnt to zero.
                                            */
                                            if (zd.zoomCnt < 1)
                                            {
                                                zd.curDpiAutoDelta = true;
                                                zd.mstrDpiAutoDelta = true;
                                            }
                                        }

                                        zd.zoomCnt -= 1;

                                        // Set this for current page so the zoom reset buttons can be updated.
                                        if (curPage == page)
                                        {
                                            curPageZoomDelta = true;
                                        }
                                    }
                                    else
                                    {
                                        // No scroll bars so continue on to next page.
                                        continue;
                                    }
                                }

                                break;

                            case ("ZoomIn"):

                                /*
                                 * Each page gets zoomed in and has it's own zoom count.
                                 * Doing it this way since we have documents with landscape
                                 * orientation mixed in with regular pages, legal, etc.
                                 * which throws a curve into handling cases where one page
                                 * gets zoomed out at zoomCnt of -2 [no scroll bars] while
                                 * another page with a different orientation is at zoomCnt
                                 * of -4 and still has scrollbars.
                                */
                                int zoomMask = 0;

                                pa.GetPgZoomDataRef((int)page - 1, ref zd);

                                if (zd == null)
                                {
                                    MessageBox.Show(string.Format("pa.GetPgZoomDataRef({0}, ref zd) has returned a null PgZoomData\n\nPage will not be Zoomed In.", page - 1), "ScaleRects(ZoomIn)", MessageBoxButtons.OK, MessageBoxIcon.Error);
                                    continue;
                                }

                                if (zd.zoomCnt < 0)
                                {
                                    if (zd.zoomCnt < -1)
                                    {
                                        zoomMask = 1;
                                        zd.curDpiAutoDelta = false;
                                    }
                                    else if (zd.mstrDpiAutoDelta)
                                    {
                                        /*
                                         * We have previously zoomed out to the point of
                                         * causing no scroll bars to appear, forcing the 
                                         * dpi to a full page or full width value which
                                         * now 'breaks' our zoom out and zoom in
                                         * relationship. That is, up until this point you
                                         * could multiply by 0.8 and then get your value back
                                         * by multiplying that result by 1.25. We lost that
                                         * ability when we zoomed out to the point of a
                                         * dpiAutoDelta which set the values based on panel
                                         * size.
                                         * 
                                         * We are now at a zoomCnt of -1 and are about to increase
                                         * it to zero, which for this page means we are back at
                                         * a point of no zoom in or out. We will use our stored 
                                         * zoom data to restore the values to the valid
                                         * 'ground zero' amounts based on the current panel size.
                                        */
                                        zoomMask = 2;
                                        zd.mstrDpiAutoDelta = false;
                                        zd.curDpiAutoDelta = false;
                                    }
                                    else
                                    {
                                        zoomMask = 1;
                                    }
                                }
                                else if (zd.zoomCnt < zoomCntMax)
                                {
                                    zoomMask = 1;
                                }

                                if (zoomMask > 0)
                                {
                                    if (zoomMask == 1)
                                    {
                                        //Regular zoom in:
                                        pgBmapH = bm.pgBmapH * zoomInVal;
                                        pgBmapW = bm.pgBmapW * zoomInVal;

                                        dpi = ps.dpi * zoomInVal;
                                    }
                                    else
                                    {
                                        // Zoom in using page zoom reset data values:
                                        pgBmapH = zd.pgBmapH;
                                        pgBmapW = zd.pgBmapW;

                                        dpi = zd.dpi;
                                    }

                                    zd.zoomCnt += 1;

                                    // Set this for current page so the zoom reset buttons can be updated.
                                    if (curPage == page)
                                    {
                                        curPageZoomDelta = true;
                                    }
                                }
                                else
                                {
                                    // Not zooming, so on to the next page.
                                    continue;
                                }

                                break;

                            case ("ZoomReset"):

                                /*
                                 * Each page will be reset back to the original dpi and bitmap
                                 * height and width values.
                                */
                                pa.GetPgZoomDataRef((int)page - 1, ref zd);

                                if (zd == null)
                                {
                                    MessageBox.Show(string.Format("pa.GetPgZoomDataRef({0}, ref zd) has returned a null PgZoomData\n\nAborting Zoom Reset for this Page.", page - 1), "ScaleRects(ZoomReset)", MessageBoxButtons.OK, MessageBoxIcon.Error);
                                    continue;
                                }

                                pgBmapH = zd.pgBmapH;
                                pgBmapW = zd.pgBmapW;
                                dpi = zd.dpi;

                                zd.zoomCnt = 0;
                                zd.mstrDpiAutoDelta = false;
                                zd.curDpiAutoDelta = false;

                                break;

                            case ("Resize"):

                                /*
                                 * We 'grow' the image for this page either vertically or horizontally
                                 * by comparing the PicPanel width/height values to the stored bitmap values.
                                 * 
                                 * We grow based on the smallest difference. We do not want visible form panel
                                 * area on ALL sides of the picture box. We recalculate the dpi and bitmap
                                 * width/height values for this page. If this page has boxes, we also make
                                 * new rectangles based on the new dpi and bitmap sizes.
                                 * 
                                 * All changed info is stored.
                                 * 
                                */
                                if ((double)(panel.Width - 2) > bm.pgBmapW)
                                {
                                    wDiff = (double)(panel.Width - 2) - bm.pgBmapW;

                                    if ((double)(panel.Height - 2) > bm.pgBmapH)
                                    {
                                        // At this point we know there will be visible panel area
                                        // around the picture box, so we need to 'grow' the image.
                                        hDiff = (double)(panel.Height - 2) - bm.pgBmapH;

                                        if (wDiff < hDiff)
                                        {
                                            // Perform "FullWidth" for this page:
                                            pgBmapW = (double)(panel.Width - 2);
                                            pgBmapH = (pgBmapW / stfPageW) * stfPageH;

                                            dpi = 72.0 * (pgBmapW / stfPageW);
                                        }
                                        else
                                        {
                                            // Perform "FullPage" for this page:
                                            pgBmapH = (double)(panel.Height - 2);
                                            pgBmapW = (pgBmapH / stfPageH) * stfPageW;

                                            dpi = 72.0 * (pgBmapH / stfPageH);
                                        }

                                        /*
                                         * Signal the zoom in and out functionality that we need
                                         * to first initialize the zoom data dictionary with
                                         * new dpi and page bitmap settings to reflect the
                                         * new image size.
                                        */
                                        initZoomReset = true;

                                        // Set this for current page so the zoom reset buttons can be disabled.
                                        if (curPage == page)
                                        {
                                            curPageZoomDelta = true;
                                        }
                                    }
                                    else
                                    {
                                        continue;
                                    }
                                }
                                else
                                {
                                    continue;
                                }

                                break;
                        }

                        pa.SetPgBMapData((int)(page - 1), pgBmapH, pgBmapW);

                        pa.GetFieldArrayRef((int)(page - 1), ref fa);

                        if (fa == null)
                        {
                            // No boxes, so just store DPI for this page.
                            pa.SetPageScale((int)(page - 1), dpi, 0, 0);
                        }
                        else
                        {
                            xScale = (pgBmapW / stfPageW);
                            yScale = (pgBmapH / stfPageH);

                            pa.SetPageScale((int)(page - 1), dpi, xScale, yScale);

                            for (int i = 0; i < fa.Count; i++)
                            {
                                Field field = fa[i];

                                x = (float)((double)field.bbox.left * xScale);
                                y = (float)((double)(stfPageH - field.bbox.bottom - (field.bbox.top - field.bbox.bottom)) * yScale);
                                width = (float)((double)(field.bbox.right - field.bbox.left) * xScale);
                                height = (float)((double)(field.bbox.top - field.bbox.bottom) * yScale);

                                field.sbox = new RectangleF(x, y, width, height);
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                    success = false;
                    MessageBox.Show(ex.ToString(), "ScaleRects():", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                finally
                {
                    if (hPdf != null)
                    {
                        STFPdf.ClosePdf(hPdf);
                    }

                    if (mode == "OpenFile")
                    {
                        if (success)
                        {
                            initZoomReset = true;
                        }
                    }
                }
            }

            return success;
        }

        // Currently not using. Mouse click event handler is doing it.
        //public Boolean PointSel(Point mousePt, ref PageArray pa)
        //{
        //    Boolean isSelected = false;
        //    List<Field> fa;

        //    fa = pa.GetFieldArray((int)(curPage - 1));

        //    if (fa != null)
        //    {
        //        for (int i = 0; i < fa.Count; i++ )
        //        {
        //            if (fa[i].sbox.Contains(mousePt))
        //            {
        //                //MessageBox.Show(string.Format("{0} was clicked!", fa[i].ref_name));

        //                // For now, we unselect selected boxes and vice versa. Both lines function as expected.
        //                //fa[i].isSelected = true;
        //                fa[i].isSelected = !fa[i].isSelected;

        //                isSelected = true;

        //                break;
        //            }
        //        }

        //    }

        //    return isSelected;
        //}

        public Boolean CheckFullWidthPage(ref double pgBmapW, ref double pgBmapH, ref double dpi, double stfPageW, double stfPageH)
        {
            double wDiff, hDiff;
            bool result = false;

            /*
             * This checks if the panel that houses the picture box is larger on all sides
             * so that you would see form area all around the image. If true, we determine
             * whether to increase the image's width or height by choosing the shortest distance.
            */
            if ((double)(panel.Width - 2) > pgBmapW)
            {
                wDiff = (double)(panel.Width - 2) - pgBmapW;

                if ((double)(panel.Height - 2) > pgBmapH)
                {
                    /*
                     * At this point we know there will be visible panel area
                     * around the picture box, so we need to 'grow' the image.
                    */
                    hDiff = (double)(panel.Height - 2) - pgBmapH;

                    if (wDiff < hDiff)
                    {
                        // Perform "FullWidth" for this page:
                        pgBmapW = (double)(panel.Width - 2);
                        pgBmapH = (pgBmapW / stfPageW) * stfPageH;
                        dpi = 72.0 * (pgBmapW / stfPageW);
                    }
                    else
                    {
                        // Perform "FullPage" for this page:
                        pgBmapH = (double)(panel.Height - 2);
                        pgBmapW = (pgBmapH / stfPageH) * stfPageW;
                        dpi = 72.0 * (pgBmapH / stfPageH);
                    }

                    result = true;
                }
            }

            return result;
        }

        public Boolean GetXmlData(ref PageArray pa)
        {
            Boolean success = false;

            if (File.Exists(xmlFile))
            {
                try
                {
                    pa = new PageArray();
                    byte[] xml = File.ReadAllBytes(xmlFile);

                    if (xml != null)
                    {
                        XmlParser parser = new XmlParser(ref pa, xml);
                        success = parser.Parse();
                    }
                }
                catch (Exception ex)
                {
                    success = false;
                    MessageBox.Show(ex.ToString(), "GetXmlData():", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }

            return success;
        }

        public Boolean ExtractFiles(string zipFile)
        {
            string tmpPath = null;
            string ZIP_PW = "5e$ACre7ru";
            Boolean retValue = false;

            tmpPath = GetTempDirectory();

            if (!string.IsNullOrEmpty(tmpPath))
            {
                if (Directory.Exists(tmpPath))
                {
                    try
                    {
                        using (ZipFile zip = ZipFile.Read(zipFile))
                        {
                            foreach (ZipEntry e in zip)
                            {
                                if (e.FileName.EndsWith(".pdf") || e.FileName.EndsWith(".xml"))
                                {
                                    if (e.UsesEncryption == true)
                                    {
                                        e.ExtractWithPassword(tmpPath, ZIP_PW);
                                    }
                                    else
                                    {
                                        e.Extract(tmpPath);
                                    }
                                }
                            }
                        }

                        pdfFile = Path.Combine(tmpPath, "doc.pdf");
                        xmlFile = Path.Combine(tmpPath, "doc.xml");

                        // It makes no sense to continue with this document if these are missing.
                        if (File.Exists(pdfFile) && File.Exists(xmlFile))
                        {
                            retValue = true;
                            //MessageBox.Show(string.Format("Extracted To: {0}", tmpPath));
                        }
                        else
                        {
                            pdfFile = null;
                            xmlFile = null;
                        }
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show(ex.ToString(), "ExtractFiles():", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }
            }

            _extPath = tmpPath;
            return retValue;
        }

        private string GetTempDirectory()
        {
            string path = null;

            try
            {
                path = Path.GetRandomFileName();
                path = Path.Combine(Path.GetTempPath(), path);
                Directory.CreateDirectory(path);
            }
            catch (Exception ex)
            {
                path = null;
                MessageBox.Show(ex.ToString(), "GetTempDirectory():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            return path;
        }

        public void CleanupFiles()
        {
            //MessageBox.Show(string.Format("Entered CleanupFiles() -> unzipPath: {0}", _extPath));

            if (!string.IsNullOrEmpty(_extPath))
            {
                if (Directory.Exists(_extPath))
                {
                    try
                    {
                        Directory.Delete(_extPath, true);
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show(ex.ToString(), string.Format("CleanupFiles({0}):", _extPath), MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }
                _extPath = null;
            }
        }

        public void CenterPicPanel()
        {
            Point center = new Point(0, 0);
            int panelW, panelH;

            if (resetScrollBars)
            {
                // Strange, but this resets these scrollbars correctly.
                panel.VerticalScroll.Value = 0;
                panel.VerticalScroll.Value = 0;

                panel.HorizontalScroll.Value = 0;
                panel.HorizontalScroll.Value = 0;
            }
            else
            {
                center = pictureBox.Location;
            }

            /*
             * Need to calculate actual panel height and width because scrollbars are problematic,
             * somtimes .ClientSize.Height and .ClientSize.Width indicate presence of scrollbars
             * when they are not really there, throwing off the centering, along with my sanity.
            */
            panelH = (bMap.Width > (panel.Width - 2) ? panel.Height - (2 + _horScrollH) : panel.Height - 2);
            panelW = (bMap.Height > (panel.Height - 2) ? panel.Width - (2 + _vertScrollW) : panel.Width - 2);

            if (panelH > bMap.Height)
            {
                center.Y = (int)(((float)panelH - (float)bMap.Height) / 2.0f);
            }

            if (panelW > bMap.Width)
            {
                center.X = (int)(((float)panelW - (float)bMap.Width) / 2.0f);
            }

            pictureBox.Location = center;
        }

        public void ClearBmaps()
        {
            if (pictureBox.Image != null)
            {
                pictureBox.Image.Dispose();
                pictureBox.Image = null;

                // Removing any scrollbars.
                pictureBox.Size = new Size(5, 5);
            }

            if (bMap != null)
            {
                bMap.Dispose();
            }

            if (_bMapBase != null)
            {
                _bMapBase.Dispose();
            }
        }

        public void SetZoomResets(ref ProcessPdf pdf, ref PageArray pa)
        {
            int page;
            PageScale ps = null;
            PgBMapData bm = null;

            /*
             * Fall back to these settings when a zoom reset occurs.
             * Most of the time theses values should be the same per page,
             * but we also have documents with mixtures of landscape and
             * legal thrown in with the 'regular' pages, so decided to this 
             * on a per page basis.
             */
            for (page = 0; page <= numPages - 1; page++)
            {
                // Get the PageScale info for the current page:
                ps = pa.GetPageScale(page);

                if (ps == null)
                {
                    MessageBox.Show(string.Format("pa.GetPageScale({0}) has returned a null ScaleField.", page), "SetZoomResets()", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    continue;
                }

                // Get the Page Bitmap Data for the current page:
                bm = pa.GetPgBMapData(page);

                if (bm == null)
                {
                    MessageBox.Show(string.Format("pa.GetPgBMapData({0}) has returned a null PgBMapData.", page), "SetZoomResets()", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    continue;
                }

                pa.SetPgZoomData(page, ps.dpi, bm.pgBmapH, bm.pgBmapW);
            }

            initZoomReset = false;
        }

    }
}
